"""
Functions and classes common between MDV grid and radar files.

"""

# Code is adapted from Nitin Bharadwaj's Matlab code

import struct
import bz2
import gzip
import zlib
from io import BytesIO
import datetime

import numpy as np

from ..core.transforms import antenna_to_cartesian

# mapping from MDV name space to CF-Radial name space
MDV_METADATA_MAP = {'instrument_name': 'data_set_source',
                    'source': 'data_set_info'}

# Information about the MDV file structure
MDV_CHUNK_INFO_LEN = 480
MDV_INFO_LEN = 512
MDV_LONG_FIELD_LEN = 64
MDV_MAX_PROJ_PARAMS = 8
MDV_MAX_VLEVELS = 122
MDV_NAME_LEN = 128
MDV_SHORT_FIELD_LEN = 16
MDV_TRANSFORM_LEN = 16
MDV_UNITS_LEN = 16
MDV_N_COORD_LABELS = 3
MDV_COORD_UNITS_LEN = 32

# (x,y) in degrees. Simple latitude-longitude grid.
# Also known as the Simple Cylindrical or Platte Carree projection.
PROJ_LATLON = 0
# (x,y) in km. Lambert Conformal Conic projection.
PROJ_LAMBERT_CONF = 3
# (x,y) in km. Polar Stereographic projection.
PROJ_POLAR_STEREO = 5
# Cartesian, (x,y) in km. This is a simple line-of-sight
# projection used for single radar sites. The formal name is
# Oblique Lambert Azimuthal projection.
PROJ_FLAT = 8
# radar data in native Plan Position Indicator (PPI) coordinates of
# range, azimuth angle and elevation angle. x is radial range (km),
# y is azimuth angle (deg), z is elev angle (deg).
PROJ_POLAR_RADAR = 9
# (x,y) in km. Oblique Stereographic projection.
PROJ_OBLIQUE_STEREO = 12
# radar data in native Range Height Indicator (RHI) coordinates.
# x is radial range (km), y is elev angle (deg), z is az angle (deg).
PROJ_RHI_RADAR = 13
#  ***************** COMPRESSION *******************
COMPRESSION_NONE = 0  # no compression
COMPRESSION_ZLIB = 3  # Lempel-Ziv
COMPRESSION_BZIP = 4  # bzip2
COMPRESSION_GZIP = 5  # Lempel-Ziv in gzip format

#  ***************** COMPRESSION CODE *******************
TA_NOT_COMPRESSED = 0x2f2f2f2f
GZIP_COMPRESSED = 0xf7f7f7f7
GZIP_NOT_COMPRESSED = 0xf8f8f8f8
BZIP_COMPRESSED = 0xf3f3f3f3
BZIP_NOT_COMPRESSED = 0xf4f4f4f4
ZLIB_COMPRESSED = 0xf5f5f5f5
ZLIB_NOT_COMPRESSED = 0xf6f6f6f6
RL8_COMPRESSION = 0xfe0103fd

#  ***************** TRANSFORM *******************
DATA_TRANSFORM_NONE = 0  # None
DATA_TRANSFORM_LOG = 1  # Natural log

#  ***************** BIT ENCODING *******************
ENCODING_INT8 = 1  # unsigned 8 bit integer
ENCODING_INT16 = 2  # unsigned 16 bit integer
ENCODING_FLOAT32 = 5  # 32 bit IEEE floating point

#  ***************** CHUNK HEADER and DATA *******************
CHUNK_DSRADAR_PARAMS = 3
CHUNK_DSRADAR_ELEVATIONS = 7
CHUNK_DSRADAR_CALIB = 10
DS_LABEL_LEN = 40
NCHAR_DS_RADAR_PARAMS = 2 * DS_LABEL_LEN
DS_RADAR_CALIB_NAME_LEN = 16
DS_RADAR_CALIB_MISSING = -9999.0


class MdvFile(object):
    """
    A file object for MDV data.

    A `MdvFile` object stores metadata and data from a MDV file. Metadata is
    stored in dictionaries as attributes of the object, field data is
    stored as NumPy ndarrays as attributes with the field name. By default
    only metadata is read initially and field data must be read using the
    `read_a_field` or `read_all_fields` methods. This behavior can be changed
    by setting the `read_fields` parameter to True.

    Parameters
    ----------
    filename : str, file-like or None.
        Name of MDV file to read or file-like object pointing to the
        beginning of such a file. None can be used to initalize an object
        which can be used for writing mdv files.
    debug : bool
        True to print out debugging information, False to supress.
    read_fields : bool
        True to read all field during initalization, False (default) only
        reads metadata.

    Notes
    -----
    This class is not stable enough for general purpose MDV reading/writing,
    nor is that the intention, but with care it can provide sufficient
    read/write capacity.

    """
    # formats for use in the struct lib
    # mapper are used to convert tuples to dics, the formats is as follows:
    # (var_name, initial_position_in_tuple, final_position_in_tuple)
    master_header_fmt = b'>28i 8i i 5i 6f 3f 12f 512s 128s 128s i'
    master_header_mapper = [
        ("record_len1", 0, 1),
        ("struct_id", 1, 2),
        ("revision_number", 2, 3),
        ("time_gen", 3, 4),
        ("user_time", 4, 5),
        ("time_begin", 5, 6),
        ("time_end", 6, 7),
        ("time_centroid", 7, 8),
        ("time_expire", 8, 9),
        ("num_data_times", 9, 10),
        ("index_number", 10, 11),
        ("data_dimension", 11, 12),
        ("data_collection_type", 12, 13),
        ("user_data", 13, 14),
        ("native_vlevel_type", 14, 15),
        ("vlevel_type", 15, 16),
        ("vlevel_included", 16, 17),
        ("grid_orientation", 17, 18),
        ("data_ordering", 18, 19),
        ("nfields", 19, 20),
        ("max_nx", 20, 21),
        ("max_ny", 21, 22),
        ("max_nz", 22, 23),
        ("nchunks", 23, 24),
        ("field_hdr_offset", 24, 25),
        ("vlevel_hdr_offset", 25, 26),
        ("chunk_hdr_offset", 26, 27),
        ("field_grids_differ", 27, 28),
        ("user_data_si328", 28, 36),
        ("time_written", 36, 37),
        ("unused_si325", 37, 42),
        ("user_data_fl326", 42, 48),
        ("sensor_lon", 48, 49),
        ("sensor_lat", 49, 50),
        ("sensor_alt", 50, 51),
        ("unused_fl3212", 51, 63),
        ("data_set_info", 63, 64),
        ("data_set_name", 64, 65),
        ("data_set_source", 65, 66),
        ("record_len2", 66, 67)
    ]

    field_header_fmt = '>17i 10i 9i 4i f f 8f 12f 4f 5f 64s 16s 16s 16s 16s i'
    field_header_mapper = [
        ("record_len1", 0, 1),
        ("struct_id", 1, 2),
        ("field_code", 2, 3),
        ("user_time1", 3, 4),
        ("forecast_delta", 4, 5),
        ("user_time2", 5, 6),
        ("user_time3", 6, 7),
        ("forecast_time", 7, 8),
        ("user_time4", 8, 9),
        ("nx", 9, 10),
        ("ny", 10, 11),
        ("nz", 11, 12),
        ("proj_type", 12, 13),
        ("encoding_type", 13, 14),
        ("data_element_nbytes", 14, 15),
        ("field_data_offset", 15, 16),
        ("volume_size", 16, 17),
        ("user_data_si32", 17, 27),
        ("compression_type", 27, 28),
        ("transform_type", 28, 29),
        ("scaling_type", 29, 30),
        ("native_vlevel_type", 30, 31),
        ("vlevel_type", 31, 32),
        ("dz_constant", 32, 33),
        ("data_dimension", 33, 34),
        ("zoom_clipped", 34, 35),
        ("zoom_no_overlap", 35, 36),
        ("unused_si32", 36, 40),
        ("proj_origin_lat", 40, 41),
        ("proj_origin_lon", 41, 42),
        ("proj_param", 42, 50),
        ("vert_reference", 50, 51),
        ("grid_dx", 51, 52),
        ("grid_dy", 52, 53),
        ("grid_dz", 53, 54),
        ("grid_minx", 54, 55),
        ("grid_miny", 55, 56),
        ("grid_minz", 56, 57),
        ("scale", 57, 58),
        ("bias", 58, 59),
        ("bad_data_value", 59, 60),
        ("missing_data_value", 60, 61),
        ("proj_rotation", 61, 62),
        ("user_data_fl32", 62, 66),
        ("min_value", 66, 67),
        ("max_value", 67, 68),
        ("min_value_orig_vol", 68, 69),
        ("max_value_orig_vol", 69, 70),
        ("unused_fl32", 70, 71),
        ("field_name_long", 71, 72),
        ("field_name", 72, 73),
        ("units", 73, 74),
        ("transform", 74, 75),
        ("unused_char", 75, 76),
        ("record_len2", 76, 77)
    ]

    vlevel_header_fmt = '>i i 122i 4i 122f 5f i'
    vlevel_header_mapper = [
        ("record_len1", 0, 1),
        ("struct_id", 1, 2),
        ("type", 2, 124),
        ("unused_si32", 124, 128),
        ("level", 128, 250),
        ("unused_fl32", 250, 255),
        ("record_len2", 255, 256)
    ]

    chunk_header_fmt = '>5i 2i 480s i'
    chunk_header_mapper = [
        ("record_len1", 0, 1),
        ("struct_id", 1, 2),
        ("chunk_id", 2, 3),
        ("chunk_data_offset", 3, 4),
        ("size", 4, 5),
        ("unused_si32", 5, 7),
        ("info", 7, 8),
        ("record_len2", 8, 9)
    ]

    compression_info_fmt = '>I I I I 2I'
    compression_info_mapper = [
        ("magic_cookie", 0, 1),
        ("nbytes_uncompressed", 1, 2),
        ("nbytes_compressed", 2, 3),
        ("nbytes_coded", 3, 4),
        ("spare", 4, 6)
    ]

    radar_info_fmt = '>12i 2i 22f 4f 40s 40s'
    radar_info_mapper = [
        ("radar_id", 0, 1),
        ("radar_type", 1, 2),
        ("nfields", 2, 3),
        ("ngates", 3, 4),
        ("samples_per_beam", 4, 5),
        ("scan_type", 5, 6),
        ("scan_mode", 6, 7),
        ("nfields_current", 7, 8),
        ("field_flag", 8, 9),
        ("polarization", 9, 10),
        ("follow_mode", 10, 11),
        ("prf_mode", 11, 12),
        ("spare_ints", 12, 14),
        ("radar_constant", 14, 15),
        ("altitude_km", 15, 16),
        ("latitude_deg", 16, 17),
        ("longitude_deg", 17, 18),
        ("gate_spacing_km", 18, 19),
        ("start_range_km", 19, 20),
        ("horiz_beam_width_deg", 20, 21),
        ("vert_beam_width_deg", 21, 22),
        ("pulse_width_us", 22, 23),
        ("prf_hz", 23, 24),
        ("wavelength_cm", 24, 25),
        ("xmit_peak_pwr_watts", 25, 26),
        ("receiver_mds_dbm", 26, 27),
        ("receiver_gain_db", 27, 28),
        ("antenna_gain_db", 28, 29),
        ("system_gain_db", 29, 30),
        ("unambig_vel_mps", 30, 31),
        ("unambig_range_km", 31, 32),
        ("measXmitPowerDbmH_dbm", 32, 33),
        ("measXmitPowerDbmV_dbm", 33, 34),
        ("prt_s", 34, 35),
        ("prt2_s", 35, 36),
        ("spare_floats", 36, 40),
        ("radar_name", 40, 41),
        ("scan_type_name", 41, 42)
    ]

    calib_fmt = '>16s 6i 51f 14f'
    calib_mapper = [
        ("radar_name", 0, 1),
        ("year", 1, 2),
        ("month", 2, 3),
        ("day", 3, 4),
        ("hour", 4, 5),
        ("minute", 5, 6),
        ("second", 6, 7),
        ("wavelength_cm", 7, 8),
        ("beamwidth_h_deg", 8, 9),
        ("beamwidth_v_deg", 9, 10),
        ("antenna_gain_h_db", 10, 11),
        ("antenna_gain_v_db", 11, 12),
        ("pulse_width_us", 12, 13),
        ("xmit_power_h_dbm", 13, 14),
        ("xmit_power_v_dbm", 14, 15),
        ("twoway_waveguide_loss_h_db", 15, 16),
        ("twoway_waveguide_loss_v_db", 16, 17),
        ("twoway_radome_loss_h_db", 17, 18),
        ("twoway_radome_loss_v_db", 18, 19),
        ("filter_loss_db", 19, 20),
        ("radar_constant_h_db", 20, 21),
        ("radar_constant_v_db", 21, 22),
        ("noise_h_co_dbm", 22, 23),
        ("noise_h_cx_dbm", 23, 24),
        ("noise_v_co_dbm", 24, 25),
        ("noise_v_cx_dbm", 25, 26),
        ("rx_gain_h_co_dbm", 26, 27),
        ("rx_gain_h_cx_dbm", 27, 28),
        ("rx_gain_v_co_dbm", 28, 29),
        ("rx_gain_v_cx_dbm", 29, 30),
        ("zh1km_co_dbz", 30, 31),
        ("zh1km_cx_dbz", 31, 32),
        ("zv1km_co_dbz", 32, 33),
        ("zv1km_cx_dbz", 33, 34),
        ("sun_h_co_dbm", 34, 35),
        ("sun_h_cx_dbm", 35, 36),
        ("sun_v_co_dbm", 36, 37),
        ("sun_v_cx_dbm", 37, 38),
        ("noise_source_h_dbm", 38, 39),
        ("noise_source_v_dbm", 39, 40),
        ("power_meas_loss_h_db", 40, 41),
        ("power_meas_loss_v_db", 41, 42),
        ("coupler_fwd_loss_h_db", 42, 43),
        ("coupler_fwd_loss_v_db", 43, 44),
        ("zdr_bias_db", 44, 45),
        ("ldr_h_bias_db", 45, 46),
        ("ldr_v_bias_db", 46, 47),
        ("system_phidp_deg", 47, 48),
        ("test_pulse_h_dbm", 48, 49),
        ("test_pulse_v_dbm", 49, 50),
        ("rx_slope_h_co_db", 50, 51),
        ("rx_slope_h_cx_db", 51, 52),
        ("rx_slope_v_co_db", 52, 53),
        ("rx_slope_v_cx_db", 53, 54),
        ("I0_h_co_dbm", 54, 55),
        ("I0_h_cx_dbm", 55, 56),
        ("I0_v_co_dbm", 56, 57),
        ("I0_v_cx_dbm", 57, 58),
        ("spare", 58, 72)
    ]

    def __init__(self, filename, debug=False, read_fields=False):
        """ initalize """
        if debug:
            print("Opening file for reading: ", filename)
        if filename is None:
            # will creat empty structures, to be filled and written later
            self.fileptr = None
        elif hasattr(filename, 'read'):
            self.fileptr = filename
        else:
            self.fileptr = open(filename, 'rb')

        if debug:
            print("Getting master header")
        self.master_header = self._get_master_header()

        if debug:
            print("getting field headers")
        nfields = self.master_header['nfields']
        self.field_headers = self._get_field_headers(nfields)

        if debug:
            print("getting vlevel headers")
        self.vlevel_headers = self._get_vlevel_headers(nfields)

        if debug:
            print("getting chunk headers")
        nchunks = self.master_header['nchunks']
        self.chunk_headers = self._get_chunk_headers(nchunks)

        if debug:
            print("Getting Chunk Data")
        # store raw chunk data, used for unknown chunk information
        self.chunk_data = [None] * self.master_header['nchunks']
        self.radar_info, self.elevations, self.calib_info = self._get_chunks(
            debug)

        if self.master_header['nfields'] > 0:
            projections = {
                PROJ_LATLON: 'latlon',
                PROJ_LAMBERT_CONF: 'lambert_conform',
                PROJ_POLAR_STEREO: 'polar_stereographic',
                PROJ_FLAT: 'flat',
                PROJ_POLAR_RADAR: 'ppi',
                PROJ_OBLIQUE_STEREO: 'oblique_stereographic',
                PROJ_RHI_RADAR: 'rhi',
            }
            self.projection = projections[self.field_headers[0]['proj_type']]

        if debug:
            print("Making usable time objects")
        self.times = self._make_time_dict()

        if debug:
            print("indexing fields")
        self.fields = self._make_fields_list()

        self.fields_data = [None] * self.master_header["nfields"]

        if read_fields:
            if debug:
                print("Reading all fields")
            self.read_all_fields()
        return

    ##################
    # public methods #
    ##################
    def write(self, filename, debug=False):
        """
        Write object data to a MDV file.

        Note that the file is not explicitly closes, use x.close() to
        close file object when complete.

        Parameters
        ----------
        filename : str or file-like
            Filename or open file object to which data will be written.
        debug : bool, options
            True to print out debugging information, False to supress.

        """
        if debug:
            print("Opening file for writing:", filename)
        if hasattr(filename, 'write'):
            self.fileptr = filename
        else:
            self.fileptr = open(filename, 'wb')
        file_start = self.fileptr.tell()

        # write fields and chunk so that offsets can be calculated
        # headers are initially zeros
        headers_size = (1024 + (416 + 1024) * self.master_header["nfields"] +
                        512 * self.master_header["nchunks"])
        self.fileptr.write(b"\x00" * headers_size)

        if debug:
            print("Writing Fields Data")
        for ifield in range(self.master_header["nfields"]):
            self._write_a_field(ifield)

        # write chunks
        if debug:
            print("Writing Chunk Data")
        self._write_chunks(debug)

        # calculate offsets
        self._calc_file_offsets()
        self.fileptr.seek(file_start)

        # write headers
        if debug:
            print("Writing master header")
        self._write_master_header()

        if debug:
            print("Writing field headers")
        self._write_field_headers(self.master_header["nfields"])

        if debug:
            print("Writing vlevel headers")
        self._write_vlevel_headers(self.master_header["nfields"])

        if debug:
            print("Writing chunk headers")
        self._write_chunk_headers(self.master_header["nchunks"])

    def read_a_field(self, fnum, debug=False):
        """
        Read a field from the MDV file.

        Parameters
        ----------
        fnum : int
            Field number to read.
        debug : bool
            True to print debugging information, False to supress.

        Returns
        -------
        field_data : array
            Field data.  This data is also stored as a object attribute under
            the field name.

        See Also
        --------
        read_all_fields : Read all fields in the MDV file.

        """

        field_header = self.field_headers[fnum]
        # if the field has already been read, return it
        if self.fields_data[fnum] is not None:
            if debug:
                print("Getting data from the object.")
            return self.fields_data[fnum]

        # field has not yet been read, populate the object and return
        if debug:
            print("No data found in object, populating")

        nz = field_header['nz']
        ny = field_header['ny']
        nx = field_header['nx']

        # read the header
        field_data = np.zeros([nz, ny, nx], dtype='float32')
        self.fileptr.seek(field_header['field_data_offset'])
        self._get_levels_info(nz)  # dict not used, but need to seek.

        for sw in range(nz):
            if debug:
                print("doing levels ", sw)

            # get the compressed level data
            compr_info = self._get_compression_info()
            if compr_info['magic_cookie'] == 0xfe0103fd:
                # Run length encoding only has 20 bytes of compression
                # information with slightly different order, back up
                # 4 bytes to read all of the compressed data.
                self.fileptr.seek(-4, 1)
                compr_data = self.fileptr.read(compr_info['spare'][0])
            else:
                compr_data = self.fileptr.read(compr_info['nbytes_coded'])
            encoding_type = field_header['encoding_type']
            if encoding_type == ENCODING_INT8:
                fmt = '>%iB' % (nx * ny)
                np_form = '>B'
            elif encoding_type == ENCODING_INT16:
                fmt = '>%iH' % (nx * ny)
                np_form = '>H'
            elif encoding_type == ENCODING_FLOAT32:
                fmt = '>%if' % (nx * ny)
                np_form = '>f'
            else:
                raise NotImplementedError('encoding: ', encoding_type)

            # decompress the level data
            if compr_info['magic_cookie'] == GZIP_COMPRESSED:
                cd_fobj = BytesIO(compr_data)
                gzip_file_handle = gzip.GzipFile(fileobj=cd_fobj)
                decompr_data = gzip_file_handle.read(struct.calcsize(fmt))
                gzip_file_handle.close()
            elif compr_info['magic_cookie'] == ZLIB_COMPRESSED:
                decompr_data = zlib.decompress(compr_data)
            elif compr_info['magic_cookie'] == BZIP_COMPRESSED:
                decompr_data = bz2.decompress(compr_data)
            elif compr_info['magic_cookie'] == TA_NOT_COMPRESSED:
                decompr_data = compr_data
            elif compr_info['magic_cookie'] == GZIP_NOT_COMPRESSED:
                decompr_data = compr_data
            elif compr_info['magic_cookie'] == ZLIB_NOT_COMPRESSED:
                decompr_data = compr_data
            elif compr_info['magic_cookie'] == BZIP_NOT_COMPRESSED:
                decompr_data = compr_data
            elif compr_info['magic_cookie'] == RL8_COMPRESSION:
                # Run length encoding of 8-bit data
                # Compression info is in a different order, namely
                # int32 : RL8_FLAG (0xfe0103fd)
                # int32 : key
                # int32 : nbytes_array (bytes of encoded data with header)
                # int32 : nbytes_full (bytes of unencoded data, no header)
                # int32 : nbytes_coded (bytes of encoded data, no header)
                key = compr_info['nbytes_uncompressed']
                decompr_size = compr_info['nbytes_coded']
                decompr_data = _decode_rle8(compr_data, key, decompr_size)
            else:
                raise NotImplementedError('unknown compression mode')

            # read the decompressed data, reshape and mask
            sw_data = np.frombuffer(decompr_data, np_form).astype('float32')
            sw_data.shape = (ny, nx)
            mask = sw_data == field_header['bad_data_value']
            np.putmask(sw_data, mask, [np.NaN])

            # scale and offset the data, store in field_data
            scale = field_header['scale']
            bias = field_header['bias']
            field_data[sw, :, :] = sw_data * scale + bias

        # store data as object attribute and return
        self.fields_data[fnum] = field_data
        return field_data

    def read_all_fields(self):
        """ Read all fields, storing data to field name attributes. """
        for i in range(self.master_header['nfields']):
            self.read_a_field(i)

    def close(self):
        """ Close the MDV file. """
        self.fileptr.close()

    ###################
    # private methods #
    ###################

    def _write_a_field(self, fnum):
        """ write field number 'fnum' to mdv file. """
        # the file pointer must be set at the correct location prior to call
        field_header = self.field_headers[fnum]
        if field_header['compression_type'] != 3:
            import warnings
            warnings.warn(
                "compression_type not implemented, converting to zlib")
            field_header['compression_type'] = 3

        field_data = self.fields_data[fnum]
        nz = field_header['nz']
        # save file position
        field_start = self.fileptr.tell()
        # write zeros to vlevel_offsets and vlevel_nbytes, these will
        # replaced by the correct data later
        self.fileptr.write(b"\x00" * 4 * 2 * nz)
        field_size = 0
        vlevel_offsets = [0] * nz
        vlevel_nbytes = [0] * nz
        for sw in range(nz):
            vlevel_offsets[sw] = field_size
            # apply scaling, offset and masking to field data
            scale = field_header['scale']
            bias = field_header['bias']
            sw_data = (field_data[sw, :, :] - bias) / scale
            if hasattr(sw_data, 'mask'):
                sw_data = np.where(
                    sw_data.mask, field_header['bad_data_value'], sw_data)

            # encode field data to the correct type, round when necessary
            encoding_type = field_header['encoding_type']
            if encoding_type == ENCODING_INT8:
                sw_data = np.round(sw_data).astype(np.uint8)
                np_form = '>B'
            elif encoding_type == ENCODING_INT16:
                sw_data = np.round(sw_data).astype(np.uint16)
                np_form = '>H'
            elif encoding_type == ENCODING_FLOAT32:
                sw_data = sw_data.astype(np.float32)
                np_form = '>f'
            else:
                raise NotImplementedError('encoding: ', encoding_type)
            uncompr_data = np.array(sw_data, dtype=np_form).tostring()
            compr_data = zlib.compress(uncompr_data)
            if len(compr_data) > len(uncompr_data):
                magic = 0xf6f6f6f6
                compr_data = uncompr_data
            else:
                magic = 0xf5f5f5f5
            compr_info = {
                'magic_cookie': magic,
                'nbytes_uncompressed': len(uncompr_data),
                'nbytes_compressed': len(compr_data) + 24,
                'nbytes_coded': len(compr_data),
                'spare': [0, 0],
            }

            self._write_compression_info(compr_info)
            self.fileptr.write(compr_data)
            field_size = field_size + len(compr_data) + 24
            vlevel_nbytes[sw] = len(compr_data) + 24
        # rewrite vlevel_offsets and vlevel_nbytes with corrected data
        field_end = self.fileptr.tell()
        self.fileptr.seek(field_start)
        vlevels_dic = {'vlevel_offsets': vlevel_offsets,
                       'vlevel_nbytes': vlevel_nbytes}
        self._write_levels_info(nz, vlevels_dic)
        self.fileptr.seek(field_end)
        field_header["volume_size"] = field_size + 2 * 4 * nz

    # get_ methods for reading headers

    def _unpack_mapped_tuple(self, l, mapper):
        """ Create a dictionary from a tuple using a mapper. """
        d = {}
        for item in mapper:
            if item[2] == item[1] + 1:
                d[item[0]] = l[item[1]]
            else:
                d[item[0]] = l[item[1]:item[2]]
            if isinstance(d[item[0]], bytes):
                d[item[0]] = d[item[0]].decode('ascii').split('\x00', 1)[0]
        return d

    def _pack_mapped(self, d, mapper, fmt):
        """ Create a packed string using a mapper and format. """
        l = [0] * mapper[-1][2]
        for item in mapper:
            if item[2] == item[1] + 1:
                l[item[1]] = d[item[0]]
                if hasattr(l[item[1]], 'encode'):   # encode str/unicode
                    l[item[1]] = l[item[1]].encode('ascii')
            else:
                l[item[1]:item[2]] = d[item[0]]
        # cast to string as Python < 2.7.7 pack does not except unicode
        fmt = str(fmt)
        return struct.pack(fmt, *l)

    def _get_master_header(self):
        """ Read the MDV master header, return a dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.master_header_mapper[-1][2]
            l[0] = 1016
            l[1] = 14142
            l[2] = 1
            l[9] = 1
            l[16] = 1
            l[17] = 1
            l[63] = ""
            l[64] = ""
            l[65] = ""
            l[66] = 1016
        else:
            l = struct.unpack(
                self.master_header_fmt,
                self.fileptr.read(struct.calcsize(self.master_header_fmt)))
        return self._unpack_mapped_tuple(l, self.master_header_mapper)

    def _write_master_header(self):
        """ Write the MDV master header. """
        # the file pointer must be set at the correct location prior to call
        d = self.master_header
        l = [0] * self.master_header_mapper[-1][2]
        for item in self.master_header_mapper:
            if item[2] == item[1] + 1:
                l[item[1]] = d[item[0]]
                if hasattr(l[item[1]], 'encode'):   # encode str/unicode
                    l[item[1]] = l[item[1]].encode('ascii')
            else:
                l[item[1]:item[2]] = d[item[0]]
        string = struct.pack(self.master_header_fmt, *l)
        self.fileptr.write(string)

    def _get_field_headers(self, nfields):
        """ Read nfields field headers, return a list of dicts. """
        # the file pointer must be set at the correct location prior to call
        return [self._get_field_header() for i in range(nfields)]

    def _write_field_headers(self, nfields):
        """ Write nfields field headers. """
        # the file pointer must be set at the correct location prior to call
        for i in range(nfields):
            self._write_field_header(self.field_headers[i])

    def _get_field_header(self):
        """ Read a single field header, return a dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.field_header_mapper[-1][2]
            l[0] = 408
            l[1] = 14143
            l[57] = 1  # scale
            l[71] = ""
            l[72] = ""
            l[73] = ""
            l[74] = ""
            l[75] = ""
            l[76] = 408
        else:
            l = struct.unpack(
                self.field_header_fmt,
                self.fileptr.read(struct.calcsize(self.field_header_fmt)))
        return self._unpack_mapped_tuple(l, self.field_header_mapper)

    def _write_field_header(self, d):
        """ Write the a single field header. """
        # the file pointer must be set at the correct location prior to call
        string = self._pack_mapped(
            d, self.field_header_mapper, self.field_header_fmt)
        self.fileptr.write(string)

    def _get_vlevel_headers(self, nfields):
        """ Read nfields vlevel headers, return a list of dicts. """
        # the file pointer must be set at the correct location prior to call
        return [self._get_vlevel_header() for i in range(nfields)]

    def _write_vlevel_headers(self, nfields):
        """ Write nfields vlevel headers. """
        # the file pointer must be set at the correct location prior to call
        for i in range(nfields):
            self._write_vlevel_header(self.vlevel_headers[i])

    def _get_vlevel_header(self):
        """ Read a single vlevel header, return a dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.vlevel_header_mapper[-1][2]
            l[0] = 1016
            l[1] = 14144
            l[255] = 1016
        else:
            l = struct.unpack(
                self.vlevel_header_fmt,
                self.fileptr.read(struct.calcsize(self.vlevel_header_fmt)))
        return self._unpack_mapped_tuple(l, self.vlevel_header_mapper)

    def _write_vlevel_header(self, d):
        """  Write the a single vfield header. """
        # the file pointer must be set at the correct location prior to call
        string = self._pack_mapped(
            d, self.vlevel_header_mapper, self.vlevel_header_fmt)
        self.fileptr.write(string)

    def _get_chunk_headers(self, nchunks):
        """ Get nchunk chunk headers, return a list of dicts. """
        # the file pointer must be set at the correct location prior to call
        return [self._get_chunk_header() for i in range(nchunks)]

    def _write_chunk_headers(self, nchunks):
        """ Write nchunk chunk headers. """
        # the file pointer must be set at the correct location prior to call
        for i in range(nchunks):
            self._write_chunk_header(self.chunk_headers[i])

    def _get_chunk_header(self):
        """ Get a single chunk header, return a dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.chunk_header_mapper[-1][2]
            l[0] = 504
            l[1] = 14145
            l[7] = ""
            l[8] = 504
        else:
            l = struct.unpack(
                self.chunk_header_fmt,
                self.fileptr.read(struct.calcsize(self.chunk_header_fmt)))
        return self._unpack_mapped_tuple(l, self.chunk_header_mapper)

    def _write_chunk_header(self, d):
        """  Write the a single chunk header. """
        # the file pointer must be set at the correct location prior to call
        string = self._pack_mapped(
            d, self.chunk_header_mapper, self.chunk_header_fmt)
        self.fileptr.write(string)

    def _get_chunks(self, debug=False):
        """ Get data in chunks, return radar_info, elevations, calib_info. """
        # the file pointer must be set at the correct location prior to call
        radar_info, elevations, calib_info = None, [], None
        for cnum, curr_chunk_header in enumerate(self.chunk_headers):

            chunk_id = curr_chunk_header['chunk_id']
            self.fileptr.seek(curr_chunk_header['chunk_data_offset'])

            if chunk_id == CHUNK_DSRADAR_PARAMS:
                if debug:
                    print('Getting radar info')
                radar_info = self._get_radar_info()

            elif chunk_id == CHUNK_DSRADAR_ELEVATIONS:
                if debug:
                    print('getting elevations')
                elevations = self._get_elevs(curr_chunk_header['size'])

            elif chunk_id == CHUNK_DSRADAR_CALIB:
                if debug:
                    print('getting cal')
                calib_info = self._get_calib()

            else:
                if debug:
                    print('getting unknown chunk %i' % chunk_id)
                self.chunk_data[cnum] = self._get_unknown_chunk(cnum)

        return radar_info, elevations, calib_info

    def _write_chunks(self, debug=False):
        """ write chunks data """
        # the file pointer must be set at the correct location prior to call
        for cnum, curr_chunk_header in enumerate(self.chunk_headers):
            chunk_id = curr_chunk_header['chunk_id']

            if chunk_id == CHUNK_DSRADAR_PARAMS:
                if debug:
                    print('writing radar info')
                self._write_radar_info(self.radar_info)

            elif chunk_id == CHUNK_DSRADAR_ELEVATIONS:
                if debug:
                    print('writing elevations')
                self._write_elevs(self.elevations)

            elif chunk_id == CHUNK_DSRADAR_CALIB:
                if debug:
                    print('writing cal')
                self._write_calib(self.calib_info)

            else:
                if debug:
                    print('writing unknown chunk %i' % chunk_id)
                self._write_unknown_chunk(self.chunk_data[cnum])

    def _get_radar_info(self):
        """ Get the radar information, return dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.radar_info_mapper[-1][2]
            l[40] = ""
            l[41] = ""
        else:
            l = struct.unpack(
                self.radar_info_fmt,
                self.fileptr.read(struct.calcsize(self.radar_info_fmt)))
        return self._unpack_mapped_tuple(l, self.radar_info_mapper)

    def _write_radar_info(self, d):
        """  Write radar information. """
        # the file pointer must be set at the correct location prior to call
        string = self._pack_mapped(
            d, self.radar_info_mapper, self.radar_info_fmt)
        self.fileptr.write(string)

    def _get_elevs(self, nbytes):
        """ Return an array of elevation read from current file position. """
        # the file pointer must be set at the correct location prior to call
        SIZE_FLOAT = 4.0
        nelevations = np.floor(nbytes / SIZE_FLOAT)
        fmt = '>%df' % (nelevations)
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        return np.array(l)

    def _write_elevs(self, l):
        """ Write an array of elevation. """
        # the file pointer must be set at the correct location prior to call
        fmt = '>%df' % (len(l))
        # cast to string as Python < 2.7.7 pack does not except unicode
        fmt = str(fmt)
        string = struct.pack(fmt, *l)
        self.fileptr.write(string)

    def _get_calib(self):
        """ Get the calibration information, return a dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.calib_mapper[-1][2]
            l[0] = ""
        else:
            l = struct.unpack(
                self.calib_fmt,
                self.fileptr.read(struct.calcsize(self.calib_fmt)))
        return self._unpack_mapped_tuple(l, self.calib_mapper)

    def _write_calib(self, d):
        """  Write calibration information. """
        # the file pointer must be set at the correct location prior to call
        string = self._pack_mapped(
            d, self.calib_mapper, self.calib_fmt)
        self.fileptr.write(string)

    def _get_compression_info(self):
        """ Get compression infomation, return a dict. """
        # the file pointer must be set at the correct location prior to call
        if self.fileptr is None:
            l = [0] * self.compression_info_mapper[-1][2]
        else:
            l = struct.unpack(
                self.compression_info_fmt,
                self.fileptr.read(struct.calcsize(self.compression_info_fmt)))
        return self._unpack_mapped_tuple(l, self.compression_info_mapper)

    def _write_compression_info(self, d):
        """ Write compression infomation. """
        # the file pointer must be set at the correct location prior to call
        string = self._pack_mapped(
            d, self.compression_info_mapper, self.compression_info_fmt)
        self.fileptr.write(string)

    def _get_unknown_chunk(self, cnum):
        """ Get raw data from chunk. """
        # the file pointer must be set at the correct location prior to call
        size = self.chunk_headers[cnum]['size']
        return self.fileptr.read(size)

    def _write_unknown_chunk(self, data):
        """ Write raw data from chunk. """
        # the file pointer must be set at the correct location prior to call
        self.fileptr.write(data)

    def _get_levels_info(self, nlevels):
        """ Get nlevel information, return a dict. """
        # the file pointer must be set at the correct location prior to call
        fmt = '>%iI %iI' % (nlevels, nlevels)
        if self.fileptr:
            l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        else:
            l = [0] * 2 * nlevels
        d = {}
        d['vlevel_offsets'] = l[:nlevels]
        d['vlevel_nbytes'] = l[nlevels: 2 * nlevels]
        return d

    def _write_levels_info(self, nlevels, d):
        """ write levels information, return a dict. """
        # the file pointer must be set at the correct location prior to call
        fmt = '%iI %iI' % (nlevels, nlevels)
        l = d['vlevel_offsets'] + d['vlevel_nbytes']
        # cast to string as Python < 2.7.7 pack does not except unicode
        fmt = str(fmt)
        string = struct.pack(fmt, *l)
        self.fileptr.write(string)

    def _calc_file_offsets(self):
        """ Calculate file offsets. """
        self.master_header["field_hdr_offset"] = 1024
        self.master_header["vlevel_hdr_offset"] = (
            1024 + 416 * self.master_header["nfields"])
        self.master_header["chunk_hdr_offset"] = (
            1024 + (416 + 1024) * self.master_header["nfields"])

        file_pos = (self.master_header["chunk_hdr_offset"] +
                    512 * self.master_header["nchunks"])
        for i in range(self.master_header["nfields"]):
            self.field_headers[i]["field_data_offset"] = file_pos
            file_pos = file_pos + self.field_headers[i]["volume_size"]

        for i in range(self.master_header["nchunks"]):
            self.chunk_headers[i]["chunk_data_offset"] = file_pos
            file_pos = file_pos + self.chunk_headers[i]["size"]

    def _make_time_dict(self):
        """ Return a time dictionary. """
        t_base = datetime.datetime(1970, 1, 1, 00, 00)
        tb = datetime.timedelta(seconds=self.master_header['time_begin'])
        te = datetime.timedelta(seconds=self.master_header['time_end'])
        tc = datetime.timedelta(seconds=self.master_header['time_centroid'])
        return {'time_begin': t_base + tb, 'time_end': t_base + te,
                'time_centroid': t_base + tc}

    def _time_dict_into_header(self):
        """ Complete time information in master_header from the time dict. """
        self.master_header['time_begin'] = self._secs_since_epoch(
            self.times['time_begin'])
        self.master_header['time_end'] = self._secs_since_epoch(
            self.times['time_end'])
        self.master_header['time_centroid'] = self._secs_since_epoch(
            self.times['time_centroid'])

    def _secs_since_epoch(self, dt):
        """ Return the number of seconds since the epoch for a datetime. """
        epoch = datetime.datetime(1970, 1, 1, 0, 0)
        td = dt - epoch
        # use td.total_seconds() in Python 2.7+
        return int((td.microseconds + (td.seconds + td.days * 24 * 3600) *
                    10**6) / 10**6)

    # misc. methods
    # XXX move some where else, there are not general mdv operations
    def _calc_geometry(self):
        """ Calculate geometry, return az_deg, range_km, el_deg. """
        nsweeps = self.master_header['max_nz']
        nrays = self.master_header['max_ny']
        ngates = self.master_header['max_nx']
        grid_minx = self.field_headers[0]['grid_minx']
        grid_miny = self.field_headers[0]['grid_miny']
        grid_dx = self.field_headers[0]['grid_dx']
        grid_dy = self.field_headers[0]['grid_dy']

        range_km = grid_minx + np.arange(ngates) * grid_dx

        if self.field_headers[0]['proj_type'] == PROJ_RHI_RADAR:
            el_deg = grid_miny + np.arange(nrays) * grid_dy
            az_deg = self.vlevel_headers[0]['level'][0:nsweeps]
        elif self.field_headers[0]['proj_type'] == PROJ_POLAR_RADAR:
            az_deg = grid_miny + np.arange(nrays) * grid_dy
            el_deg = self.vlevel_headers[0]['level'][0:nsweeps]
        else:
            proj_type = self.field_headers[0]['proj_type']
            message = ("Unsupported projection type: %i, " % (proj_type) +
                       "is MDV file in antenna coordinates?")
            raise NotImplementedError(message)

        return az_deg, range_km, el_deg

    def _make_carts_dict(self):
        """ Return a carts dictionary, distances in meters. """
        az_deg, range_km, el_deg = self._calc_geometry()
        # simple calculation involving 4/3 earth radius
        nsweeps = self.master_header['max_nz']
        nrays = self.master_header['max_ny']
        ngates = self.master_header['max_nx']
        xx = np.empty([nsweeps, nrays, ngates], dtype=np.float32)
        yy = np.empty([nsweeps, nrays, ngates], dtype=np.float32)
        zz = np.empty([nsweeps, nrays, ngates], dtype=np.float32)

        if self.projection == 'rhi':
            rg, ele = np.meshgrid(range_km, el_deg)
            rg = np.array(rg, dtype=np.float64)
            ele = np.array(ele, dtype=np.float64)
            for aznum in range(nsweeps):
                azg = np.ones(rg.shape, dtype=np.float64) * az_deg[aznum]
                x, y, z = antenna_to_cartesian(rg, azg, ele)
                zz[aznum, :, :] = z
                xx[aznum, :, :] = x
                yy[aznum, :, :] = y

        elif self.projection == 'ppi':
            rg, azg = np.meshgrid(range_km, az_deg)
            rg = np.array(rg, dtype=np.float64)
            azg = np.array(azg, dtype=np.float64)
            for elnum in range(nsweeps):
                ele = np.ones(rg.shape, dtype=np.float64) * el_deg[elnum]
                x, y, z = antenna_to_cartesian(rg, azg, ele)
                zz[elnum, :, :] = z
                xx[elnum, :, :] = x
                yy[elnum, :, :] = y

        return {'x': xx, 'y': yy, 'z': zz}

    def _make_fields_list(self):
        """ Return a list of fields. """
        fh = self.field_headers
        return [fh[i]['field_name'] for i in range(len(fh))]


def _decode_rle8(compr_data, key, decompr_size):
    """ Decode 8-bit MDV run length encoding. """
    # Encoding is described in section 7 of:
    # http://rap.ucar.edu/projects/IHL/RalHtml/protocols/mdv_file/
    # This function would benefit greate by being rewritten in Cython
    data = np.frombuffer(compr_data, dtype='>B')
    out = np.empty((decompr_size, ), dtype='uint8')
    data_ptr = 0
    out_ptr = 0
    while data_ptr != len(data):
        v = data[data_ptr]
        if v != key:  # not encoded
            out[out_ptr] = v
            data_ptr += 1
            out_ptr += 1
        else:   # run length encoded
            count = data[data_ptr+1]
            value = data[data_ptr+2]
            out[out_ptr:out_ptr+count] = value
            data_ptr += 3
            out_ptr += count
    return out.tostring()


class _MdvVolumeDataExtractor(object):
    """
    Class facilitating on demand extraction of data from a MDV file.

    Parameters
    ----------
    mdvfile : MdvFile
        Open MdvFile object to extract data from.
    field_num : int
        Field number of data to be extracted.
    fillvalue : int
        Value used to fill masked values in the returned array.
    two_dims : bool.
        True to combine the first and second dimension of the array when
        returning the data, False will return a three dimensional array.

    """

    def __init__(self, mdvfile, field_num, fillvalue, two_dims=True):
        """ initialize the object. """
        self.mdvfile = mdvfile
        self.field_num = field_num
        self.fillvalue = fillvalue
        self.two_dims = two_dims

    def __call__(self):
        """ Return an array containing data from the referenced volume. """
        # grab data from MDV object, mask and reshape
        data = self.mdvfile.read_a_field(self.field_num)
        data[np.where(np.isnan(data))] = self.fillvalue
        data[np.where(data == 131072)] = self.fillvalue
        data = np.ma.masked_equal(data, self.fillvalue)
        if self.two_dims:
            data.shape = (data.shape[0] * data.shape[1], data.shape[2])
        return data



import argparse

import matplotlib.pyplot as plt
import pyart

if __name__ == '__main__':

    # parse the arguments
    parser = argparse.ArgumentParser(
        description='Plot a field from a radar file.')

    # positional arguments
    parser.add_argument('filename', type=str, help='name of radar file')
    parser.add_argument('field', type=str, help='field in radar to plot')
    parser.add_argument('figname', type=str, nargs='?',
                        help=('file to save figure as, if not specified '
                              'figure is displayed on screen'))

    # value optional arguments
    parser.add_argument('--tilt', type=int, default=0,
                        help=('radar tilt/sweep to plot. ' 'A value of 0, '
                              'the first tilt, is used if not specified.'))
    parser.add_argument('--vmin', type=float, default=None,
                        help='luminance minimum value')
    parser.add_argument('--vmax', type=float, default=None,
                        help='luminance maximum value')
    parser.add_argument('--cmap', type=str, default='jet',
                        help='name of colormap used to plot data')
    parser.add_argument('--title', type=str, default=None,
                        help='title of the plot')
    parser.add_argument('--axis_labels', type=str, nargs=2,
                        default=(None, None),
                        metavar=('XAXIS_LABEL', 'YAXIS_LABEL'),
                        help='x-axis and y-axis labels')
    parser.add_argument('--colorbar_label', type=str, default=None,
                        help='colorbar label')

    # boolean optional arguments
    parser.add_argument('--no_mask_outside', action='store_true',
                        help='do not mask data outside of vmin and vmax')
    parser.add_argument('--no_title', action='store_true',
                        help='do not title the figure')
    parser.add_argument('--no_axis_labels', action='store_true',
                        help='do not label the axes')
    parser.add_argument('--no_colorbar', action='store_true',
                        help='do not create a colorbar')

    parser.add_argument('-v', '--version', action='version',
                        version='Py-ART version %s' % (pyart.__version__))

    # ingest arguments
    igroup = parser.add_argument_group(
        title='ingest method, optional',
        description=('The method of file ingest can be specified. '
                     'If no ingest is specified, the format of the file will '
                     'be used to determine the best ingest method. '
                     'Specify only one of the following:'))

    igroup.add_argument('--sigmet', action='store_true',
                        help='Sigmet/IRIS ingest')
    igroup.add_argument('--mdv', action='store_true', help='MDV ingest')
    igroup.add_argument('--cfradial', action='store_true',
                        help='CF/Radial ingest')
    igroup.add_argument('--rsl', action='store_true',
                        help='RSL ingest')
    igroup.add_argument('--nexrad_archive', action='store_true',
                        help='NEXRAD level 2 archive ingest')
    igroup.add_argument('--nexrad_cdm', action='store_true',
                        help='NEXRAD level 2 CDM ingest')

    args = parser.parse_args()

    # read in the file
    if args.sigmet:
        radar = pyart.io.read_sigmet(args.filename)
    elif args.mdv:
        radar = pyart.io.read_mdv(args.filename)
    elif args.cfradial:
        radar = pyart.io.read_cfradial(args.filename)
    elif args.rsl:
        radar = pyart.io.read_rsl(args.filename)
    elif args.nexrad_archive:
        radar = pyart.io.read_nexrad_archive(args.filename)
    elif args.nexrad_cdm:
        radar = pyart.io.read_nexrad_cdm(args.filename)
    else:
        radar = pyart.io.read(args.filename)

    # create the plot
    fig = plt.figure()
    ax = fig.add_subplot(111)

    if args.field not in radar.fields:
        print("Valid fields:")
        for f in radar.fields:
            print('\t' + f)
        raise ValueError('invalid field: ' + args.field)

    display = pyart.graph.RadarDisplay(radar)
    if radar.scan_type == 'ppi':
        display.plot_ppi(
            args.field, args.tilt, ax=ax, fig=fig,
            vmin=args.vmin, vmax=args.vmax, cmap=args.cmap,
            mask_outside=(not args.no_mask_outside),
            title_flag=(not args.no_title),
            axislabels_flag=(not args.no_axis_labels),
            colorbar_flag=(not args.no_colorbar),
            title=args.title,
            axislabels=args.axis_labels,
            colorbar_label=args.colorbar_label)
    elif radar.scan_type == 'rhi':
        display.plot_rhi(
            args.field, args.tilt, ax=ax, fig=fig,
            vmin=args.vmin, vmax=args.vmax, cmap=args.cmap,
            mask_outside=(not args.no_mask_outside),
            title_flag=(not args.no_title),
            axislabels_flag=(not args.no_axis_labels),
            colorbar_flag=(not args.no_colorbar),
            title=args.title,
            axislabels=args.axis_labels,
            colorbar_label=args.colorbar_label)
    elif radar.scan_type == 'vpt':
         display.plot_rhi(
            args.field, ax=ax, fig=fig,
            vmin=args.vmin, vmax=args.vmax, cmap=args.cmap,
            mask_outside=(not args.no_mask_outside),
            title_flag=(not args.no_title),
            axislabels_flag=(not args.no_axis_labels),
            colorbar_flag=(not args.no_colorbar),
            title=args.title,
            axislabels=args.axis_labels,
            colorbar_label=args.colorbar_label)
    else:
        raise ValueError('Cannot plot radar scan_type: ' + radar.scan_type)

    # show the plot if no figname provided, otherwise save the figure
    if args.figname is None:
        plt.show()
    else:
        fig.savefig(args.figname)



""" Unit tests for the circular_stats.py module. """

from pyart.util import circular_stats

import numpy as np
from numpy.testing import assert_almost_equal


def test_mean_of_two_angles():
    mean = circular_stats.mean_of_two_angles(np.pi/4 + 0.2, 7 * np.pi / 4)
    assert_almost_equal(mean, 0.10, 2)


def test_mean_of_two_angles_deg():
    mean = circular_stats.mean_of_two_angles_deg(359.9, 0.5)
    assert_almost_equal(mean, 0.20, 2)


def test_angles_radians():
    dist = [2*np.pi-0.1, 0, 0.1]
    mean = circular_stats.angular_mean(dist)
    assert_almost_equal(mean, 0., 2)
    std = circular_stats.angular_std(dist)
    assert_almost_equal(std, 0.08, 2)


def test_angles_degrees():
    dist = [350, 0, 10, 20, 30]
    mean = circular_stats.angular_mean_deg(dist)
    assert_almost_equal(mean, 10., 0)
    std = circular_stats.angular_std_deg(dist)
    assert_almost_equal(std, 14., 0)


def test_angles_interval():
    dist = [7.5, 8.5, 9.5, 9.5, -9.5, -8.5]
    mean = circular_stats.interval_mean(dist, -10, 10)
    assert_almost_equal(mean, 9.5, 1)
    std = circular_stats.interval_std(dist, -10, 10)
    assert_almost_equal(std, 1.3, 1)



"""
Py-ART can act as bridge to other community software projects.

The functionality in this namespace is available in other pyart namespaces.

Current extensions:
    * wradlib https://wradlib.org/



"""

from .wradlib_bridge import texture_of_complex_phase
from .. import retrieve as _retrieve
_retrieve.texture_of_complex_phase = texture_of_complex_phase



"""
Corrects polarimetric variables for noise

"""

import numpy as np

from ..config import get_metadata, get_field_name


def correct_noise_rhohv(radar, urhohv_field=None, snr_field=None,
                        zdr_field=None, nh_field=None, nv_field=None,
                        rhohv_field=None):
    """
    Corrects RhoHV for noise according to eq. 6 in Gourley et al. 2006.
    This correction should only be performed if noise has not been subtracted
    from the signal during the moments computation.

    Parameters
    ----------
    radar : Radar
        Radar object.
    urhohv_field : str, optional
        Name of the RhoHV uncorrected for noise field.
    snr_field, zdr_field, nh_field, nv_field : str, optional
        Names of the SNR, ZDR, horizontal channel noise in dBZ and vertical
        channel noise in dBZ used to correct RhoHV.
    rhohv_field : str, optional
        Name of the rhohv field to output.

    Returns
    -------
    rhohv : dict
        Noise corrected RhoHV field.

    References
    ----------
    Gourley et al. Data Quality of the Meteo-France C-Band Polarimetric
    Radar, JAOT, 23, 1340-1356

    """
    # parse the field parameters
    if urhohv_field is None:
        urhohv_field = get_field_name('uncorrected_cross_correlation_ratio')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if nh_field is None:
        nh_field = get_field_name('noisedBZ_hh')
    if nv_field is None:
        nv_field = get_field_name('noisedBZ_vv')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')

    # extract fields from radar
    if urhohv_field in radar.fields:
        urhohv = radar.fields[urhohv_field]['data']
    else:
        raise KeyError('Field not available: ' + urhohv_field)
    if snr_field in radar.fields:
        snrdB_h = radar.fields[snr_field]['data']
    else:
        raise KeyError('Field not available: ' + snr_field)
    if zdr_field in radar.fields:
        zdrdB = radar.fields[zdr_field]['data']
    else:
        raise KeyError('Field not available: ' + zdr_field)
    if nh_field in radar.fields:
        nh = radar.fields[nh_field]['data']
    else:
        raise KeyError('Field not available: ' + nh_field)
    if nv_field in radar.fields:
        nv = radar.fields[nv_field]['data']
    else:
        raise KeyError('Field not available: ' + nv_field)

    snr_h = np.ma.power(10., 0.1*snrdB_h)
    zdr = np.ma.power(10., 0.1*zdrdB)
    alpha = np.ma.power(10., 0.1*(nh-nv))

    rhohv_data = urhohv*np.ma.sqrt((1.+1./snr_h)*(1.+zdr/(alpha*snr_h)))
    rhohv_data[rhohv_data > 1.] = 1.

    rhohv = get_metadata(rhohv_field)
    rhohv['data'] = rhohv_data

    return rhohv


def correct_bias(radar, bias=0., field_name=None):
    """
    Corrects a radar data bias. If field name is none the correction is
    applied to horizontal reflectivity by default.

    Parameters
    ----------
    radar : Radar
        Radar object.
    bias : float, optional
        The bias magnitude.
    field_name: str, optional
        Names of the field to be corrected.

    Returns
    -------
    corrected_field : dict
        The corrected field

    """
    # parse the field parameters
    if field_name is None:
        field_name = get_field_name('reflectivity')

    # extract fields from radar
    if field_name in radar.fields:
        field_data = radar.fields[field_name]['data']
    else:
        raise KeyError('Field not available: ' + field_name)

    corr_field_data = field_data - bias

    if field_name.startswith('corrected_'):
        corr_field_name = field_name
    else:
        corr_field_name = 'corrected_'+field_name

    corr_field = get_metadata(corr_field_name)
    corr_field['data'] = corr_field_data

    return corr_field



"""
Py-ART has a robust function for mapping radar data from the collected radar
coordinates to Cartesian coordinates.

"""

from .grid_mapper import map_to_grid, grid_from_radars
from .grid_mapper import example_roi_func_constant
from .grid_mapper import example_roi_func_dist
from .grid_mapper import example_roi_func_dist_beam
from .gates_to_grid import map_gates_to_grid

__all__ = [s for s in dir() if not s.startswith('_')]



# check that the dummy NEXRAD file is simlar to non-dummy file.

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose

import pyart

NEXRAD_FILE = 'KATX20130717_195021_V06'
OUTPUT_FILE = 'KATX20130717_195021_V06_DUMMY'


def test_dummy_similar():
    radar1 = pyart.io.read_nexrad_archive(NEXRAD_FILE)
    radar2 = pyart.io.read_nexrad_archive(OUTPUT_FILE)
    assert radars_similar(radar1, radar2)


def radars_similar(r1, r2):

    ###########################
    # Attribute that are None #
    ###########################
    assert dics_similar(r1.altitude_agl, r2.altitude_agl)
    assert dics_similar(r1.target_scan_rate, r2.target_scan_rate)
    assert dics_similar(r1.scan_rate, r2.scan_rate)
    assert dics_similar(r1.antenna_transition, r2.antenna_transition)
    assert dics_similar(r1.radar_calibration, r2.radar_calibration)

    #########################
    # Dictionary attributes #
    #########################
    #assert dics_similar(r1.time, r2.time)          # start time mismatch
    assert dics_similar(r1.range, r2.range)
    #assert dics_similar(r1.metadata, r2.metadata)  # DO not match

    assert dics_similar(r1.latitude, r2.latitude)
    assert dics_similar(r1.longitude, r2.longitude)
    #assert dics_similar(r1.altitude, r2.altitude)  # differ by 10 meters

    assert dics_similar(r1.sweep_number, r2.sweep_number)
    assert dics_similar(r1.sweep_mode, r2.sweep_mode)
    assert dics_similar(r1.fixed_angle, r2.fixed_angle)
    assert dics_similar(r1.sweep_start_ray_index, r2.sweep_start_ray_index)
    assert dics_similar(r1.sweep_end_ray_index, r2.sweep_end_ray_index)

    assert dics_similar(r1.azimuth, r2.azimuth)
    assert dics_similar(r1.elevation, r2.elevation)

    ###########
    # scalars #
    ###########

    assert r1.ngates == r2.ngates
    assert r1.nrays == r2.nrays
    assert r1.nsweeps == r2.nsweeps
    assert r1.scan_type == r2.scan_type

    ##########
    # fields #
    ##########

    print(r1.fields.keys())
    print(r2.fields.keys())
    assert set(r1.fields.keys()).difference(r2.fields.keys()) == set()

    for field in r1.fields:
        print(field)
        assert dics_similar(r1.fields[field], r2.fields[field])

    #radar1.fields
    return True


def dics_similar(dic1, dic2):
    """ Determine if two dictionaries are similar. """
    if dic1 is None:
        if dic2 is None:
            return True
        else:
            return False

    # all keys in both dictionaries
    print(dic1.keys())
    print(dic2.keys())
    assert set(dic1.keys()).difference(dic2.keys()) == set()

    for key in dic1.keys():
        if key == 'data':
            continue
        print(key, dic1[key], dic2[key])
        assert dic1[key] == dic2[key]

    # do not check data, since it should not match
    return True



"""
Routines for reading ARM vertically-pointing radar ingest (e.g., a1) files.
These files are characterized by being NetCDF files that do not fully conform
to the CF/Radial convention. Nonetheless this module borrows heavily from the
existing CF/Radial module.

"""

import netCDF4
import numpy as np

from ..io import cfradial
from ..config import FileMetadata
from ..core.radar import Radar


def read_kazr(filename, field_names=None, additional_metadata=None,
              file_field_names=False, exclude_fields=None,
              include_fields=None):
    """
    Read K-band ARM Zenith Radar (KAZR) NetCDF ingest data.

    Parameters
    ----------
    filename : str
        Name of NetCDF file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.

    Returns
    -------
    radar : Radar
        Radar object.
    """

    # create metadata retrieval object
    filemetadata = FileMetadata(
        'cfradial', field_names, additional_metadata, file_field_names,
        exclude_fields, include_fields)

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])
    metadata['n_gates_vary'] = 'false'

    # 4.2 Dimensions (do nothing)

    # 4.3 Global variable -> move to metadata dictionary
    if 'volume_number' in ncvars:
        metadata['volume_number'] = int(ncvars['volume_number'][:])
    else:
        metadata['volume_number'] = 0

    global_vars = {'platform_type': 'fixed', 'instrument_type': 'radar',
                   'primary_axis': 'axis_z'}
    # ignore time_* global variables, these are calculated from the time
    # variable when the file is written.
    for var, default_value in global_vars.items():
        if var in ncvars:
            metadata[var] = str(netCDF4.chartostring(ncvars[var][:]))
        else:
            metadata[var] = default_value

    # 4.4 coordinate variables -> create attribute dictionaries
    time = cfradial._ncvar_to_dict(ncvars['time'])
    _range = cfradial._ncvar_to_dict(ncvars['range'])

    # 4.5 Ray dimension variables

    # 4.6 Location variables -> create attribute dictionaries
    # the only difference in this section to cfradial.read_cfradial is the
    # minor variable name differences:
    # latitude -> lat
    # longitude -> lon
    # altitdue -> alt
    latitude = cfradial._ncvar_to_dict(ncvars['lat'])
    longitude = cfradial._ncvar_to_dict(ncvars['lon'])
    altitude = cfradial._ncvar_to_dict(ncvars['alt'])

    # 4.7 Sweep variables -> create atrribute dictionaries
    # this is the section that needed the most work since the initial NetCDF
    # file did not contain any sweep information
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.array([0], dtype=np.int32)

    sweep_mode = filemetadata('sweep_mode')
    sweep_mode['data'] = np.array(['vertical_pointing'], dtype=np.str)

    fixed_angle = filemetadata('fixed_angle')
    fixed_angle['data'] = np.array([90.0], dtype=np.float32)

    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_start_ray_index['data'] = np.array([0], dtype=np.int32)

    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_end_ray_index['data'] = np.array(
        [ncvars['time'].size - 1], dtype=np.int32)

    # first sweep mode determines scan_type
    # this module is specific to vertically-pointing data
    scan_type = 'vpt'

    # 4.8 Sensor pointing variables -> create attribute dictionaries
    # this section also required some changes since the initial NetCDF did not
    # contain any sensor pointing variables
    azimuth = filemetadata('azimuth')
    azimuth['data'] = 0.0 * np.ones(ncvars['time'].size, dtype=np.float32)

    elevation = filemetadata('elevation')
    elevation['data'] = 90.0 * np.ones(ncvars['time'].size, dtype=np.float32)

    # 4.9 Moving platform geo-reference variables

    # 4.10 Moments field data variables -> field attribute dictionary
    # all variables with dimensions of 'time', 'range' are fields
    keys = [k for k, v in ncvars.items() if
            v.dimensions == ('time', 'range')]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            if include_fields is not None and not key in include_fields:
                continue
            field_name = key
        fields[field_name] = cfradial._ncvar_to_dict(ncvars[key])

    # 4.5 instrument_parameters sub-convention -> instrument_parameters dict
    # this section needed multiple changes and/or additions since the
    # instrument parameters were primarily located in the global attributes
    # this section is likely still incomplete
    omega = float(ncobj.radar_operating_frequency.split()[0])
    frequency = filemetadata('frequency')
    frequency['data'] = np.array([omega / 1e9], dtype=np.float32)

    prt_mode = filemetadata('prt_mode')
    prt_mode['data'] = np.array(['fixed'], dtype=np.str)

    prf = float(ncobj.pulse_repetition_frequency.split()[0])
    prt = filemetadata('prt')
    prt['data'] = (1.0 / prf) * np.ones(ncvars['time'].size, dtype=np.float32)

    v_nq = float(ncobj.nyquist_velocity.split()[0])
    nyquist_velocity = filemetadata('nyquist_velocity')
    nyquist_velocity['data'] = v_nq * np.ones(ncvars['time'].size,
                                              dtype=np.float32),
    samples = int(ncobj.num_spectral_averages)
    n_samples = filemetadata('n_samples')
    n_samples['data'] = samples * np.ones(ncvars['time'].size, dtype=np.int32)

    # 4.6 radar_parameters sub-convention -> instrument_parameters dict
    # this section needed multiple changes and/or additions since the
    # radar instrument parameters were primarily located in the global
    # attributes
    # this section is likely still incomplete
    instrument_parameters = {
        'frequency': frequency,
        'prt_mode': prt_mode,
        'prt': prt,
        'nyquist_velocity': nyquist_velocity,
        'n_samples': n_samples,
    }

    # 4.7 lidar_parameters sub-convention -> skip
    # 4.8 radar_calibration sub-convention -> skip

    # close NetCDF object
    ncobj.close()

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)
